package gui;

import java.awt.Color;
import java.awt.GridLayout;
import java.awt.KeyEventDispatcher;
import java.awt.KeyboardFocusManager;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Map;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Receiver;
import javax.sound.midi.ShortMessage;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;

import receiver.MidiCommon;

public class SoundSimulator {

	private class KeyboardButtonListener implements KeyEventDispatcher{
		private char[] qwertyKeys = null;
		
		private Hashtable<Character, Boolean> pressedKeyTable = new Hashtable<Character, Boolean>();
		
		private ArrayList<ShortMessage> m = new ArrayList<ShortMessage>();
		
		public KeyboardButtonListener(char... k){
			qwertyKeys = k;
			initializeTables(k);
		}
		
		private void initializeTables(char[] keys){
			for(char k : keys){
				pressedKeyTable.put(k, false);
			}
		}
		
		public boolean dispatchKeyEvent(KeyEvent e) {
			int key;
			int velocity = 120;
			
			for(char qwertyKey : qwertyKeys){
				if(e.getKeyChar() == Character.toLowerCase(qwertyKey) 
						|| e.getKeyChar() == Character.toUpperCase(qwertyKey)){
					char pressedKey = Character.toLowerCase(e.getKeyChar());
					key = getNote(pressedKey);
					
					if(e.getID() == KeyEvent.KEY_PRESSED && !pressedKeyTable.get(pressedKey)){
						ShortMessage midiMsg = generateMidiMessage(key, velocity, true); 
						receiver.send(midiMsg, -1);
						m.add(midiMsg);
						//SoundSimulator.midiApp.update(MidiCommon.recognizeChord(m));
						stateLabelTable.get(pressedKey).setText("ON");
						stateLabelTable.get(pressedKey).setForeground(Color.DARK_GRAY);
						pressedKeyTable.put(pressedKey, true);
						
						if(SoundSimulator.midiApp != null)
							SoundSimulator.midiApp.setSoundField(MidiCommon.recognizeChordForSoundSimulator(m));
						
						//if(SoundSimulator.practiceApp != null)
							//SoundSimulator.practiceApp.update(MidiCommon.recognizeChordForSoundSimulator(m));
					}
					
					else if(e.getID() == KeyEvent.KEY_RELEASED){
						receiver.send(generateMidiMessage(key, velocity, false), -1);
						stateLabelTable.get(pressedKey).setText("OFF");
						stateLabelTable.get(pressedKey).setForeground(Color.GRAY);
						pressedKeyTable.put(pressedKey, false);
						
						if(SoundSimulator.midiApp != null){
							SoundSimulator.midiApp.update(MidiCommon.recognizeChordForSoundSimulator(m));
						}
						
						if(SoundSimulator.practiceApp != null)
							SoundSimulator.practiceApp.update(MidiCommon.recognizeChordForSoundSimulator(m));
						
						m.clear();
					}
				}
			}
			
			return false;
		}
	}
	
	public SoundSimulator(){
		setReceiver();
		initializeNoteTable();
		initializeStateTable();
		
		try{
		keyboardManager = MIDIQwertyfier.focusManager;
		}
		catch(Exception e){
			System.err.println("MIDI QWERTYfier not defined");
		}
		
		if(keyboardManager == null){
			System.out.println("FAILED TO RETRIEVE KeyboardFocusManager FROM MIDI-Application");
			keyboardManager = KeyboardFocusManager.getCurrentKeyboardFocusManager();
		}
		
		keyButtonListener =
			new KeyboardButtonListener(
					/* Assigning Keys */
						//Left hand keys
						's', 'd', 'f',
						//Space
						' ',
						//Right hand keys
						'j', 'k', 'l');
		
		keyboardManager.addKeyEventDispatcher(keyButtonListener);
		
		JPanel contentPane = new JPanel(new GridLayout(2, 6, 15, 5));

		contentPane.add(d4Label);
		contentPane.add(f4Label);
		contentPane.add(a4Label);
		contentPane.add(c5Label);
		contentPane.add(e5Label);
		contentPane.add(g5Label);
		contentPane.add(h5Label);
		
		contentPane.add(d4State);
		contentPane.add(f4State);
		contentPane.add(a4State);
		contentPane.add(c5State);
		contentPane.add(e5State);
		contentPane.add(g5State);
		contentPane.add(h5State);
		
		container = new JFrame("Sound Simulator");
		container.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				close();
				container.dispose();
			}
		});
		container.setContentPane(contentPane);
		container.pack();
		container.setLocation(300, 300);
		container.setVisible(true);
	}
	
	public SoundSimulator(MIDIQwertyfier midiApp){
		SoundSimulator.midiApp = midiApp;
		new SoundSimulator();
	}
	
	public SoundSimulator(PracticeProgram practiceApp){
		SoundSimulator.practiceApp = practiceApp;
		new SoundSimulator();
	}
	
	private ShortMessage generateMidiMessage(int key, int velocity, boolean isNote_On){
		ShortMessage midiMessage = null;
		int nChannel = 0;
		int nKey = key;
		int nVelocity = velocity;
		int nStatus = isNote_On ? ShortMessage.NOTE_ON : ShortMessage.NOTE_OFF;
		
		try
		{
			midiMessage = new ShortMessage();
			midiMessage.setMessage(nStatus, nChannel, nKey, nVelocity);
			
			//printMessageInfo(midiMessage, isNote_On);
		}
		catch (InvalidMidiDataException e)
		{
			System.err.println(e);
			//e.printStackTrace();
		}
		
		return midiMessage;
	}
	
	private void setReceiver(){
		MidiCommon.listDevicesAndExit(true, true);
		/*
		try
		{
			receiver = MidiSystem.getMidiDevice(MidiCommon.getMidiDeviceInfo(2)).getReceiver();
			System.out.println(receiver);
			return;
		}
		catch (MidiUnavailableException e){}
		*/
		try {
			receiver = MidiSystem.getReceiver();
		} catch (MidiUnavailableException e) {
			System.err.println("Was not able to retrieve Receiver");
			e.printStackTrace();
			System.exit(1);
		}
	}
	
	private void printMessageInfo(ShortMessage midi, boolean isNote_On){
		StringBuffer info = new StringBuffer();
		info.append(isNote_On ? "On Message: " : "Off Message: ");
		info.append(midi.getStatus());
		info.append(" ");
		info.append(midi.getData1());
		info.append(" ");
		info.append(midi.getData2());
		
		System.out.println(info);
	}
	
/*	private Sequencer getSequencer(){
		Sequencer seq = null;
		try {
			seq = MidiSystem.getSequencer();
		} catch (MidiUnavailableException e) {
			e.printStackTrace();
		}
		return seq;
	}*/
	
	private int getNote(char c){
		return noteTable.get(c);
	}
	
	private void initializeNoteTable(){
		noteTable = new Hashtable <Character, Integer>();
		//D4
		noteTable.put('s', 62);
		//F4
		noteTable.put('d', 65);
		//A4
		noteTable.put('f', 69);
		//C5
		noteTable.put(' ', 72);
		//E5
		noteTable.put('j', 76);
		//G5
		noteTable.put('k', 79);
		//H5 (aka A5)
		noteTable.put('l', 81);
	}
	
	private void initializeStateTable(){
		stateLabelTable = new Hashtable<Character, JLabel>();
		
		stateLabelTable.put('s', d4State);
		stateLabelTable.put('d', f4State);
		stateLabelTable.put('f', a4State);
		stateLabelTable.put(' ', c5State);
		stateLabelTable.put('j', e5State);
		stateLabelTable.put('k', g5State);
		stateLabelTable.put('l', h5State);
		
		for(Map.Entry<Character, JLabel> entry : stateLabelTable.entrySet()){
			entry.getValue().setForeground(Color.GRAY);
		}
	}
	
	public void close(){
		receiver = null;
		container.dispose();
		keyboardManager.removeKeyEventDispatcher(keyButtonListener);
	}

	//Table components
	Hashtable<Character, Integer> 	noteTable		= null;
	Hashtable<Character, JLabel>	stateLabelTable		= null;
	
	//MIDI components
	Receiver 		receiver 		= null;
	
	//GUI components	
	static 	JFrame 	container 	= null;
	private JFrame parentContaiFrame = null;
	
	JLabel 	d4State 	= 	new JLabel("OFF");
	JLabel 	f4State 	= 	new JLabel("OFF");
	JLabel 	a4State 	= 	new JLabel("OFF");
	JLabel 	e5State 	= 	new JLabel("OFF");
	JLabel 	g5State 	= 	new JLabel("OFF");
	JLabel 	h5State 	= 	new JLabel("OFF");
	JLabel	c5State		=	new JLabel("OFF");
	
	JLabel	d4Label		= 	new JLabel("D4");
	JLabel	f4Label		= 	new JLabel("F4");
	JLabel	a4Label		= 	new JLabel("A4");
	JLabel	e5Label		= 	new JLabel("E5");
	JLabel	g5Label		= 	new JLabel("G5");
	JLabel	h5Label		= 	new JLabel("H5");
	JLabel	c5Label		= 	new JLabel("C5");
	
	//MIDIQwertyfier 	midiApp 	= new MIDIQwertyfier();
	private static MIDIQwertyfier 	midiApp 				= null;
	private static PracticeProgram	practiceApp				= null;
	private static KeyboardFocusManager keyboardManager 			= null;
	private static KeyboardButtonListener keyButtonListener	= null;
	
	final SoundSimulator 	me 		= this;
	
	public static void main(String[] args){
		new SoundSimulator();
	}
}

