package translator;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.StringReader;

import org.jdom.Document;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

/**
 * This is class provides users to handle XML files in more
 * convenient manner. </br>
 * Main functionalities of this class are reading and 
 * converting XML file into Document, converting the Document
 * back to XML format and validating the XML or the Document using
 * pre-defined schema.
 */
public class ParserLib {

	/**
	 * XmLOutputter which is used to generate XML code
	 */
	public static XMLOutputter xmlOutputter = new XMLOutputter(Format
			.getPrettyFormat());

	/**
	 * Default constructor, currently does nothing
	 */
	public ParserLib() {
	}

	/**
	 * Builds and returns SAXBuilder. 
	 * @param validation Validates the XML, as per schema defined in
	 * "http://apache.org/xml/features/validation/schema", if it is true.
	 * @return <B>SAXBuilder</B> 
	 */
	private final static SAXBuilder getSaxBuilder(boolean validation) {
		SAXBuilder saxBuilder = new SAXBuilder(
				"org.apache.xerces.parsers.SAXParser", validation);

		saxBuilder.setFeature("http://xml.org/sax/features/validation",
								validation);
		// set shcema support
		saxBuilder
				.setFeature("http://apache.org/xml/features/validation/schema",
							validation);
		return saxBuilder;
	}

	/**
	 * Parse the input XML file with/without validation and output 
	 * the Document object
	 * 
	 * @param inputXMLFile Path of the XML file (in String)
	 * @param validation boolean argument which defines whether the
	 * the XML file needs to be defined or not.
	 * @return <B>Document</B> Tree representation of the XML file.
	 * @throws IOException
	 * @throws JDOMException
	 */
	public final static Document parser(String inputXMLFile, boolean validation)
			throws IOException, JDOMException {

		File inputXML = new File(inputXMLFile);
		if(!inputXML.exists()) {
			System.err.println("Specified not exists");
			return null;
		}
		else {
			return getSaxBuilder(validation).build(inputXMLFile);
		}
	}

	/**
	 * Parse the input XML file with/without validation and output 
	 * the Document object
	 * 
	 * @param inputXML StringReader for the XML file
	 * @param validation boolean argument which defines whether the
	 * the XML file needs to be defined or not.
	 * @return <B>Document</B> Tree representation of the XML file.
	 * @throws IOException
	 * @throws JDOMException
	 */
	public final static Document parser(StringReader inputXML,
			boolean validation) throws IOException, JDOMException {

		return getSaxBuilder(validation).build(inputXML);
	}

	/**
	 * Validate the Document.  It will find out elements that are not
	 * supposed to be in the Document by using Schema. 
	 * 
	 * @param doc XML in Document form
	 * @return Document that has been validated
	 * @throws JDOMException
	 * @throws IOException
	 */
	public final static Document validator(Document doc) throws JDOMException,
			IOException {

		return getSaxBuilder(true).build(
											new StringReader(xmlOutputter
													.outputString(doc)));
		// return saxBuilder.build(new
		// StringReader(xmlOutputter.outputString(doc)));
	}

	/**
	 * Validate the Document.  It will find out elements that are not
	 * supposed to be in the Document by using Schema.
	 *  
	 * @param input Path of the XML file.
	 * @return
	 * @throws JDOMException
	 * @throws IOException
	 */
	public final static Document validator(String input) throws JDOMException,
			IOException {
		return getSaxBuilder(true).build(new StringReader(input));
	}

	/**
	 * Validate the Document.  It will find out elements that are not
	 * supposed to be in the Document by using Schema.
	 * 
	 * @param strReader StringReader that references a XML file
	 * @return
	 * @throws JDOMException
	 * @throws IOException
	 */
	public final static Document validator(StringReader strReader)
			throws JDOMException, IOException {
		return getSaxBuilder(true).build(strReader);
	}

	/**
	 * Write the in-memory parsed XML file to an external file
	 * 
	 * @param outputXMLFile Path of the output file
	 * @param doc XML in Document form
	 * @throws IOException
	 */
	public final static void writeToFile(String outputXMLFile, Document doc)
			throws IOException {

		FileOutputStream outStream = new FileOutputStream(outputXMLFile);

		xmlOutputter.output(doc, outStream);
		outStream.flush();
		outStream.close();
	}
}
