package translator;


import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.Date;
import java.util.Hashtable;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Vector;

import org.jdom.Attribute;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;

/**
 * This is a class which stores the translation rules that are already defined in
 * XML file.  This class allows users to easily add, modify and remove rules.
 * The XML file represented as a Document variable and there is a Hashtable that
 * stores list of Rules with its Sound value as key value.  It is used for accessing
 * the Rule element for faster modifying and removing.  
 */
public class TranslationRule {
	private Hashtable<String, Element> soundTable = new Hashtable<String, Element>();
	private Document xmlDocument;

	/**
	 * Default constructor of the TranslationRule
	 * @param filepath Either absolute or canonical path of the XML file that stores 
	 * translation rules 
	 * @throws IOException
	 * @throws JDOMException
	 */
	public TranslationRule(String filepath) throws IOException, JDOMException{
		xmlDocument = ParserLib.parser(filepath, false);
		polish(xmlDocument);
	}
	
	public TranslationRule(Document document){
		xmlDocument = document;
		polish(xmlDocument);
	}

	/**
	 * Convert characters that have been modified when they were converted into XML<br>
	 * Also update soundTable with the Rule element of the Document received as an argument.
	 * @param d XML Document
	 */
	private void polish(Document d){
		Element root = d.getRootElement();
		for(Object o : root.getAttributes()){
			Attribute a = (Attribute) o;
			replaceConvertedChar(a);
		}

		Element translationRules = root.getChild("TranslationRules");
		for(Object o : translationRules.getChildren()){
			Element rule = (Element) o;
			Attribute actionAtt = rule.getAttribute("Action");
			Attribute soundAtt = rule.getAttribute("Sound");
			String sound = sortSequence(soundAtt.getValue());
			replaceConvertedChar(actionAtt);

			soundTable.put(sound, rule);
		}
	}
	
	public void refreshTable(){
		polish(xmlDocument);
	}

	/**
	 * Replace characters that were converted in the process of storing into XML files
	 * @param Attribute
	 */
	private void replaceConvertedChar(Attribute a){
		String converted = a.getValue()
		.replaceAll("&amp;", "&")
		.replaceAll("&lt;", "<")
		.replaceAll("&gt;", ">")
		.replaceAll("&quot;", "\"");
		a.setValue(converted);
	}
	
	/**
	 * Returns the name of the author
	 * @return name of the Author
	 */
	public String getAuthor(){
		return xmlDocument.getRootElement().getAttributeValue("Author");
	}

	/**
	 * Returns the comment written for the Rule
	 * @return comment of the TranslationRule
	 */
	public String getComment(){
		return xmlDocument.getRootElement().getAttributeValue("Comment");
	}

	/**
	 * Returns Hashtable<String, Element> that stores every Rule of the Document (with its Sound
	 * as its Key) 
	 * @return Hashtable
	 */
	public Hashtable<String, Element> getHashtable(){
		return soundTable;
	}
	
	/**
	 * Returns the Document of the XML
	 * @return Document
	 */
	public Document getDocument(){
		return xmlDocument;
	}

	/**
	 * Returns the Element that represents the Rule for the given Sound
	 * @param sound
	 * @return Element
	 */
	public Element getRule(String sound){
		return soundTable.get(sound);
	}

	/**
	 * Add a rule to the TranslationRules
	 * @param sound Sound in String
	 * @param action Action in String
	 */
	public void addRule(String sound, String action){
		Element rule = new Element("Rule");
		sound = sortSequence(sound);
		rule.setAttribute("Sound", sound);
		rule.setAttribute("Action", action);

		Element translationRules = xmlDocument.getRootElement().getChild("TranslationRules");
		translationRules.addContent(rule);
		
		soundTable.put(sound, rule);
	}

	/**
	 * Remove a rule with the given Sound
	 * @param sound Sound in String
	 * @return True if deleted; false if not.
	 */
	public boolean removeRule(String sound){
		Element rule = soundTable.remove(sortSequence(sound));
		boolean isDeleted = rule.getParentElement().removeContent(rule);
		return isDeleted;
	}

	/**
	 * Returns Action for the given Sound
	 * @param sound Sound in String
	 * @return Action in String
	 */
	public String getAction(String sound){
		Element rule = soundTable.get(sortSequence(sound));
		try{
			String actionValue = rule.getAttribute("Action").getValue();
			return actionValue;
		}
		catch(NullPointerException e){
			return "";
		}
	}

	/**
	 * Returns Sound for the given Action
	 * @param action Action in String
	 * @return Sound in String
	 */
	public String getSound(String action){
		for(Map.Entry<String, Element> m : soundTable.entrySet()){
			Element e = m.getValue();
			if(e.getAttributeValue("Action").equals(action))
				return e.getAttributeValue("Sound");
		}
		return null;
	}
	
	/**
	 * Modifies the Action for given Sound.
	 * 
	 * @param sound Sound in String
	 * @param newAction
	 * @return boolean representation of whether there was a rule with the given sound
	 * and it was successfully modified or not.
	 */
	public boolean modifyAction(String sound, String newAction){
		Element rule = soundTable.get(sortSequence(sound));
		try{
			rule.getAttribute("Action").setValue(newAction);
			return true;
		}
		catch(NullPointerException e){
			return false;
		}
	}
	
	/**
	 * Modify old sound with new sound
	 * @param oldSound oldSound in String
	 * @param newSound newSound in String
	 * @return
	 */ 
	public boolean modifySound(String oldSound, String newSound){
		try{
			Element ruleElement = soundTable.remove(sortSequence(oldSound));
			newSound = sortSequence(newSound);
			ruleElement.setAttribute("Sound", newSound);
			soundTable.put(newSound, ruleElement);
			return true;
		}
		catch(Exception e){
			return false;
		}
	}

	/**
	 * Sort the given sequence of Sounds.  By sorting sequence of sounds, 
	 * chords with different order but the same combination can give the same Action. 
	 * @param seq Sequence of notes
	 * @return Sorted sequence of notes
	 */
	public static String sortSequence(String seq){
		if(seq.equals("")) return "";
		
		String currentSequence=seq;
		for(int i = 0; i < 10; i++){
			if (currentSequence.contains(i+"")){
				currentSequence = currentSequence.replaceAll(i+"", i+" ");
			}
		}
		StringTokenizer noteT = new StringTokenizer(currentSequence);
		Vector<String> chordList = new Vector<String>();
		while(noteT.hasMoreTokens()){
			chordList.add(noteT.nextToken());
		}
		Collections.sort(chordList);
		String newSequence="";
		for(int i = 0; i <= chordList.indexOf(chordList.lastElement()); i++){
			newSequence += chordList.elementAt(i);
		}
		return newSequence;
	}
	
	public static void main (String[] args){
		System.out.println(new Date());
		test();
	}

	private static void test() {
		try{
			TranslationRule rule = new TranslationRule("./Jason.xml");
			System.out.println("Action for Sound:'C' : "+rule.getAction("C"));
			System.out.println("Action for Sound:'ZXCV' (invalid sound) : "+ rule.getAction("XZCV"));
			System.out.println("Sound for Action:'Hello' : "+rule.getSound("Hello"));
			System.out.println("Sound for Action:'Yellow' (invalid action) : "+rule.getSound("Yellow"));
			rule.addRule("BE", "My Name is Jason");
			System.out.println("Modifying Action : "+rule.modifyAction("BE", "Apple"));
			System.out.println("Modifying Invalid Action : "+rule.modifyAction("BaaE", "Apple"));
			System.out.println("Modifying Sound : "+rule.modifySound("Hello", "ASDF"));
			ParserLib.xmlOutputter.output(rule.getDocument(), System.out);
		}
		catch(Exception e){
			e.printStackTrace();
		}
	}
}
